#!/usr/bin/env python

from .isodates import (
    asUtc,
    toLocal,
    parseLocalTime,
    assertLocalDateTime,
    )

import datetime

import unittest

def localTime(string):
    isSummer = string.endswith("S")
    if isSummer: string=string[:-1]
    return parseLocalTime(string, isSummer)

class LocalTime_Test(unittest.TestCase):

    def test_localTime_fullySummer(self):
        self.assertEqual(
            str(localTime("2016-08-15 02:00:00")),
            "2016-08-15 02:00:00+02:00")

    def test_localTime_fullyWinter(self):
        self.assertEqual(
            str(localTime("2016-01-01 02:10:00")),
            "2016-01-01 02:10:00+01:00")

    def test_localTime_badTz_ignored(self):
        self.assertEqual(
            str(localTime("2016-01-01 02:00:00S")),
            "2016-01-01 02:00:00+01:00")

    def test_localTime_badSummerTz_ignored(self):
        self.assertEqual(
            str(localTime("2016-08-15 02:00:00")),
            "2016-08-15 02:00:00+02:00")

    def test_localTime_beforeOctoberChange(self):
        self.assertEqual(
            str(localTime("2016-10-30 02:00:00S")),
            "2016-10-30 02:00:00+02:00")

    def test_localTime_afterOctoberChange(self):
        self.assertEqual(
            str(localTime("2016-10-30 02:00:00")),
            "2016-10-30 02:00:00+01:00")

    def test_localTime_SIgnored(self):
        self.assertEqual(
            str(localTime("2016-10-30 03:00:00S")),
            "2016-10-30 03:00:00+01:00")

    @unittest.skip('toreview: it should fail')
    def test_localTime_unexistingHour(self):
        self.assertEqual(
            str(localTime("2016-03-27 02:00:00")),
            "2016-03-27 02:00:00+01:00")

    def test_localTime_atWinterMidnight(self):
        self.assertEqual(
            str(localTime("2016-01-01 00:00:00")),
            "2016-01-01 00:00:00+01:00")

    def test_assertLocalDateTime_withDate(self):
        with self.assertRaises(AssertionError) as ctx:
            assertLocalDateTime('myname', datetime.date(2016,1,1))
        self.assertEqual(ctx.exception.args[0],
            "myname should be a datetime")

    def test_assertLocalDateTime_withNaive(self):
        with self.assertRaises(AssertionError) as ctx:
            assertLocalDateTime('myname', datetime.datetime(2016,1,1))
        self.assertEqual(ctx.exception.args[0],
            "myname should have timezone")

    def test_assertLocalDateTime_withUTC(self):
        with self.assertRaises(AssertionError) as ctx:
            assertLocalDateTime('myname', asUtc(datetime.datetime(2016,1,1)))
        self.assertEqual(ctx.exception.args[0],
            "myname has UTC timezone")

    def test_assertLocalDateTime_withLocal(self):
        assertLocalDateTime('myname', toLocal(datetime.datetime(2016,1,1)))
        # No assert





# vim: et ts=4 sw=4



#!/usr/bin/env python

from .mongotimecurve import MongoTimeCurve
from somutils.isodates import localisodate
import os
import pymongo
from datetime import date
from . import testutils
from .resource import (
    ProductionMeter,
    ProductionPlant,
    ProductionAggregator,
    )

import unittest
import pytest

def local_file(filename):
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), filename)

class Meter_Test(unittest.TestCase):
    def setUp(self):
        self.databasename = 'generationkwh_test'
        self.collection = 'production'

        self.connection = pymongo.MongoClient()
        self.connection.drop_database(self.databasename)
        self.db = self.connection[self.databasename]
        self.curveProvider = MongoTimeCurve(self.db, self.collection)
        self.row1 = [0,0,0,0,0,0,0,0,3,6,5,4,8,17,34,12,12,5,3,1,0,0,0,0,0,]
        self.row2 = [0,0,0,0,0,0,0,0,4,7,6,5,9,18,35,13,13,6,4,2,0,0,0,0,0,]

    def tearDown(self):
        self.connection.drop_database('generationkwh_test')

    def setupEmptyMeter(self, **kwd):
        return ProductionMeter(
            1,
            'meterName',
            'meterDescription',
            True,
            curveProvider = self.curveProvider,
            **kwd
            )

    def setupMeter(self, **kwd):
        m = self.setupEmptyMeter(**kwd)
        self.curveProvider.update(
            localisodate("2015-09-04"), 'meterName', 'ae',
            self.row1+self.row2)
        return m

    def test__get_kwh__empty(self):
        m = self.setupEmptyMeter()
        self.assertEqual(
            list(m.get_kwh(
                date(2015,9,4),
                date(2015,9,5))),
            2*25*[0]
            )

    def test__get_kwh__filled(self):
        m = self.setupMeter()
        self.assertEqual(
            list(m.get_kwh(
                date(2015,9,4),
                date(2015,9,5))),
            self.row1 + self.row2)

    def test__get_kwh__filled__whenFiltered(self):
        m = self.setupMeter(first_active_date="2015-09-05")
        self.assertEqual(
            list(m.get_kwh(
                date(2015,9,4),
                date(2015,9,5))),
            [0]*25 + self.row2)

    def test__get_kwh__filled__whenFiltered_onStart(self):
        m = self.setupMeter(first_active_date="2015-09-04")
        self.assertEqual(
            list(m.get_kwh(
                date(2015,9,4),
                date(2015,9,5))),
            self.row1 + self.row2)

    def test__get_kwh__filled__whenFiltered_onEnd(self):
        m = self.setupMeter(first_active_date="2015-09-06")
        self.assertEqual(
            list(m.get_kwh(
                date(2015,9,4),
                date(2015,9,5))),
            [0]*50)

    def test_lastDate_empty(self):
        m = self.setupEmptyMeter()
        self.assertEqual(m.lastMeasurementDate(), None)

    def test_lastDate_filled(self):
        m = self.setupMeter()
        self.assertEqual(m.lastMeasurementDate(), date(2015,9,5))

    def test_firstDate_empty(self):
        m = self.setupEmptyMeter()
        self.assertEqual(m.firstMeasurementDate(), None)

    def test_firstDate_filled(self):
        m = self.setupMeter()
        self.assertEqual(m.firstMeasurementDate(), date(2015,9,4))


class Resource_Test(unittest.TestCase):

    def setUp(self):
        self.databasename = 'generationkwh_test'
        self.collection = 'production'

        self.connection = pymongo.MongoClient()
        self.connection.drop_database(self.databasename)
        self.db = self.connection[self.databasename]
        self.curveProvider = MongoTimeCurve(self.db, self.collection)
        self.row1 = [0,0,0,0,0,0,0,0,3,6,5,4,8,17,34,12,12,5,3,1,0,0,0,0,0]
        self.row2 = [0,0,0,0,0,0,0,0,4,7,6,5,9,18,35,13,13,6,4,2,0,0,0,0,0]

    def tearDown(self):
        self.connection.drop_database('generationkwh_test')

    def setupMeter(self, n, name):
        return ProductionMeter(
            id=n,
            name = name,
            description = 'meterDescription{}'.format(n),
            enabled = True,
            curveProvider = self.curveProvider,
            )

    def fillMeter(self, name, start):
        self.curveProvider.update(
            localisodate(start), name, 'ae',
            self.row1+self.row2)


    def test__get_kwh__empty(self):
        m = self.setupMeter(1, 'counter')

        p = ProductionPlant(1,'plantName','plantDescription',True, meters=[m])
        aggr = ProductionAggregator(1,'aggrName','eggrDescription',True, plants=[p])

        self.assertEqual(
            list(aggr.get_kwh(
                date(2015,9,4),
                date(2015,9,5))),
                2*25*[0] 
                )
      
    def test__get_kwh__onePlantOneMeter(self):
        m = self.setupMeter(1, '20150904')

        p = ProductionPlant(1,'plantName','plantDescription',True, meters=[m])
        aggr = ProductionAggregator(1,'aggrName','aggrDescription',True, plants=[p])
        self.fillMeter('20150904', '2015-09-04')

        self.assertEqual(
            list(aggr.get_kwh(
                date(2015,9,4),
                date(2015,9,5))),
            [
                0,0,0,0,0,0,0,0,3,6,5,4,8,17,34,12,12,5,3,1,0,0,0,0,0,
                0,0,0,0,0,0,0,0,4,7,6,5,9,18,35,13,13,6,4,2,0,0,0,0,0,
            ])

    def test__get_kwh__onePlantTwoMeters(self):
        m1 = self.setupMeter(1,'m1')
        self.fillMeter('m1', '2015-09-04')
        m2 = self.setupMeter(2,'m2')
        self.fillMeter('m2', '2015-09-04')

        p = ProductionPlant(1,'plantName','plantDescription',True, meters=[m1,m2])
        aggr = ProductionAggregator(1,'aggrName','aggrDescription',True, plants=[p])

        self.assertEqual(
            list(aggr.get_kwh(
                date(2015,9,4),
                date(2015,9,5))),
            [
                0,0,0,0,0,0,0,0,6,12,10, 8,16,34,68,24,24,10,6,2,0,0,0,0,0,
                0,0,0,0,0,0,0,0,8,14,12,10,18,36,70,26,26,12,8,4,0,0,0,0,0,
            ])

    import sys
    @pytest.mark.skipif(sys.version_info > (2,8), reason="py3 numpy requires same dimensions when sum arrays")
    def test__get_kwh__twoPlantsOneMeter(self):
        m1 = self.setupMeter(1, 'm1')
        self.fillMeter('m1', '2015-09-04')
        p1 = ProductionPlant(1,'plantName1','plantDescription1',True, meters=[m1])
        p2 = ProductionPlant(2,'plantName2','plantDescription2',True)
        aggr = ProductionAggregator(1,'aggrName','aggrDescription',True,plants=[p1,p2])

        self.assertEqual(
            list(aggr.get_kwh(
                date(2015,9,4),
                date(2015,9,5))),
            [
                0,0,0,0,0,0,0,0,3,6,5,4,8,17,34,12,12,5,3,1,0,0,0,0,0,
                0,0,0,0,0,0,0,0,4,7,6,5,9,18,35,13,13,6,4,2,0,0,0,0,0,
            ])

    def test__get_kwh__twoPlantsTwoMeters(self):
        m1 = self.setupMeter(1, 'm1')
        self.fillMeter('m1', '2015-09-04')
        p1 = ProductionPlant(1,'plantName1','plantDescription1',True, meters=[m1])

        m2 = self.setupMeter(2, 'm2')
        self.fillMeter('m2', '2015-09-04')
        p2 = ProductionPlant(2,'plantName2','plantDescription2',True, meters=[m2])

        aggr = ProductionAggregator(1,'aggrName','aggreDescription',True,
            plants=[p1, p2])

        self.assertEqual(
            list(aggr.get_kwh(
                date(2015,9,4),
                date(2015,9,5))),
            [
                0,0,0,0,0,0,0,0,6,12,10, 8,16,34,68,24,24,10,6,2,0,0,0,0,0,
                0,0,0,0,0,0,0,0,8,14,12,10,18,36,70,26,26,12,8,4,0,0,0,0,0,
            ])

    def test_lastDate_empty(self):
        m = self.setupMeter(1, '20150904')
        p = ProductionPlant(1,'plantName','plantDescription',True, meters=[m])
        aggr = ProductionAggregator(1,'aggrName','aggreDescription',True,
            plants=[p])

        self.assertEqual(aggr.lastMeasurementDate(), None)
    
    def test_firstDate_empty(self):
        m = self.setupMeter(1, '20150904')
        p = ProductionPlant(1,'plantName','plantDescription',True, meters=[m])
        aggr = ProductionAggregator(1,'aggrName','aggreDescription',True, plants=[p])

        self.assertEqual(aggr.firstMeasurementDate(), None)

    def test_lastDate_onePlantOneMeter(self):
        m = self.setupMeter(1, 'm1')
        self.fillMeter('m1', '2015-09-04')
        p = ProductionPlant(1,'plantName','plantDescription',True, meters=[m])
        aggr = ProductionAggregator(1,'aggrName','aggreDescription',True, plants=[p])

        self.assertEqual(aggr.lastMeasurementDate(), date(2015,9,5))

    def test_firstDate_onePlantOneMeter(self):
        m = self.setupMeter(1, 'm1')
        self.fillMeter('m1', '2015-09-04')
        p = ProductionPlant(1,'plantName','plantDescription',True, meters=[m])
        aggr = ProductionAggregator(1,'aggrName','aggreDescription',True, plants=[p])

        self.assertEqual(aggr.firstMeasurementDate(), date(2015,9,4))

    def test_lastDate_onePlantTwoMeters(self):
        m1 = self.setupMeter(1, 'm1')
        m2 = self.setupMeter(2, 'm2')
        self.fillMeter('m1', '2015-09-04')
        self.fillMeter('m2', '2015-08-04')

        p = ProductionPlant(1,'plantName','plantDescription',True, meters=[m1,m2])
        aggr = ProductionAggregator(1,'aggrName','aggreDescription',True, plants=[p])

        self.assertEqual(aggr.lastMeasurementDate(), date(2015,8,5))

    def test_firstDate_onePlantTwoMeters(self):
        m1 = self.setupMeter(1, 'm1')
        m2 = self.setupMeter(2, 'm2')
        self.fillMeter('m1', '2015-09-04')
        self.fillMeter('m2', '2015-08-04')

        p = ProductionPlant(1,'plantName','plantDescription',True, meters=[m1,m2])
        aggr = ProductionAggregator(1,'aggrName','aggreDescription',True, plants=[p])

        self.assertEqual(aggr.firstMeasurementDate(), date(2015,8,4))

    def test_lastDate_twoPlantsTwoMeters(self):
        m1 = self.setupMeter(1, 'm1')
        m2 = self.setupMeter(2, 'm2')
        self.fillMeter('m1', '2015-09-04')
        self.fillMeter('m2', '2015-08-04')

        p1 = ProductionPlant(1,'plantName1','plantDescription1',True, meters=[m1])
        p2 = ProductionPlant(2,'plantName2','plantDescription2',True, meters=[m2])
        aggr = ProductionAggregator(1,'aggrName','aggreDescription',True, plants=[p1,p2])

        self.assertEqual(aggr.lastMeasurementDate(), date(2015,8,5))

    def test_firstDate_twoPlantsTwoMeters(self):
        m1 = self.setupMeter(1, 'm1')
        m2 = self.setupMeter(2, 'm2')
        self.fillMeter('m1', '2015-09-04')
        self.fillMeter('m2', '2015-08-04')

        p1 = ProductionPlant(1,'plantName1','plantDescription1',True, meters=[m1])
        p2 = ProductionPlant(2,'plantName2','plantDescription2',True, meters=[m2])
        aggr = ProductionAggregator(1,'aggrName','aggreDescription',True, plants=[p1,p2])

        self.assertEqual(aggr.firstMeasurementDate(), date(2015,8,4))

class Mix_Test(unittest.TestCase):

    def setUp(self):
        self.databasename = 'generationkwh_test'
        self.collection = 'production'

        self.connection = pymongo.MongoClient()
        self.connection.drop_database(self.databasename)
        self.db = self.connection[self.databasename]
        self.curveProvider = MongoTimeCurve(self.db, self.collection)
        self.row1 = [0,0,0,0,0,0,0,0,3,6,5,4,8,17,34,12,12,5,3,1,0,0,0,0,0]
        self.row2 = [0,0,0,0,0,0,0,0,4,7,6,5,9,18,35,13,13,6,4,2,0,0,0,0,0]

    def tearDown(self):
        self.connection.drop_database('generationkwh_test')

    def setupMeter(self, n, name):
        return ProductionMeter(
            id=n,
            name = name,
            description = 'meterDescription{}'.format(n),
            enabled = True,
            curveProvider = self.curveProvider,
            )

    def fillMeter(self, name, start):
        self.curveProvider.update(
            localisodate(start), name, 'ae',
            self.row1+self.row2)


    def test_firstActiveDate_noPlants(self):
        aggr = ProductionAggregator(1,'aggrName','aggreDescription',True, plants=[])

        self.assertEqual(aggr.firstActiveDate(), None)

    def test_firstActiveDate_singlePlant(self):
        m1 = self.setupMeter(1, 'm1')
        p1 = ProductionPlant(1,'plantName1','plantDescription1',True, '2000-01-01', meters=[m1])
        aggr = ProductionAggregator(1,'aggrName','aggreDescription',True, plants=[p1])

        self.assertEqual(aggr.firstActiveDate(), date(2000,1,1))

    def test_firstActiveDate_manyPlants(self):
        m1 = self.setupMeter(1, 'm1')
        m2 = self.setupMeter(2, 'm2')
        p1 = ProductionPlant(1,'plantName1','plantDescription1',True, '2000-01-01', meters=[m1])
        p2 = ProductionPlant(2,'plantName2','plantDescription2',True, '2001-01-01', meters=[m2])
        aggr = ProductionAggregator(1,'aggrName','aggreDescription',True, plants=[p1,p2])

        self.assertEqual(aggr.firstActiveDate(), date(2000,1,1))

        


# vim: et ts=4 sw=4



# -*- coding: utf-8 -*-

import unittest
from yamlns import namespace as ns

# Readable verbose testcase listing
unittest.TestCase.__str__ = unittest.TestCase.id

def assertNsEqual(self, dict1, dict2):
    """
    Asserts that both dict have equivalent structure.
    If parameters are strings they are parsed as yaml.
    Comparation by comparing the result of turning them
    to yaml sorting the keys of any dict within the structure.
    """
    def parseIfString(nsOrString):
        if isinstance(nsOrString, dict):
            return nsOrString
        return ns.loads(nsOrString)

    def sorteddict(d):
        if type(d) in (dict, ns):
            return ns(sorted(
                (k, sorteddict(v))
                for k,v in d.items()
                ))
        if type(d) in (list, tuple):
            return [sorteddict(x) for x in d]
        return d
    dict1 = sorteddict(parseIfString(dict1))
    dict2 = sorteddict(parseIfString(dict2))

    return self.assertMultiLineEqual(dict1.dump(), dict2.dump())

def _inProduction():
    import erppeek_wst
    import dbconfig
    c = erppeek_wst.ClientWST(**dbconfig.erppeek)
    c.begin()
    destructive_testing_allowed = c._execute(
        'res.config', 'get', 'destructive_testing_allowed', False)
    c.rollback()
    c.close()

    if destructive_testing_allowed: return False
    return True

def destructiveTest(decorated):
    return unittest.skipIf(_inProduction(),
        "Destructive test being run in a production setup!!")(decorated)



# vim: ts=4 sw=4 et



import numpy as np

from somutils.isodates import (
    isodate,
    dateToLocal,
    assertDate,
    )
import datetime

"""
TODOs
+ Ordre
- Duplicats
- Gaps
+ Padding
"""

class Resource(object):
    id = None 
    name = None
    description = None
    enabled = None

    def __init__(self, id, name, description, enabled):
        self.id = id 
        self.name = name
        self.description = description
        self.enabled = enabled

class ParentResource(Resource):

    def __init__(self, id, name, description, enabled, children=[]):
        super(ParentResource, self).__init__(id, name, description, enabled)
        self.children = children

    def get_kwh(self, start, end):

        assertDate('start', start)
        assertDate('end', end)

        return np.sum([
            child.get_kwh(start, end)
            for child in self.children
            if child.enabled
            ], axis=0)

    def firstMeasurementDate(self):
        return min([
            child.firstMeasurementDate()
            for child in self.children
            if child.enabled
            ])

    def lastMeasurementDate(self):
        return min([
            child.lastMeasurementDate()
            for child in self.children
            if child.enabled
            ])

class ProductionAggregator(ParentResource):
    def __init__(self, id, name, description, enabled, plants=[]):
        super(ProductionAggregator, self).__init__(
            id, name, description, enabled, children=plants)

    def firstActiveDate(self):
        if not self.children: return None
        return isodate(min(plant.first_active_date for plant in self.children))

class ProductionPlant(ParentResource):
    def __init__(self, id, name, description, enabled, first_active_date=None, last_active_date=None, meters=[]):
        super(ProductionPlant, self).__init__(
            id, name, description, enabled, children=meters)
        self.first_active_date = first_active_date

class ProductionMeter(Resource):
    def __init__(self, *args, **kwargs):
        self.first_active_date = kwargs.pop('first_active_date', None)
        self.first_active_date = self.first_active_date and isodate(self.first_active_date)
        self.curveProvider = kwargs.pop('curveProvider', None)
        super(ProductionMeter, self).__init__(*args, **kwargs)

    def get_kwh(self, start, end):

        assertDate('start', start)
        assertDate('end', end)

        data = self.curveProvider.get(
            start=dateToLocal(start),
            stop=dateToLocal(end),
            filter=self.name,
            field='ae',
            )

        if self.first_active_date and self.first_active_date >= start:
            nbins= (self.first_active_date-start).days * 25
            data[:nbins] = 0

        return data

    def lastMeasurementDate(self):
        result = self.curveProvider.lastFullDate(self.name)
        return result and result.date()
    
    def firstMeasurementDate(self):
        result = self.curveProvider.firstFullDate(self.name)
        return result and result.date()

# vim: et ts=4 sw=4



# -*- coding: utf-8 -*-

import os
from osv import osv, fields
from mongodb_backend import osv_mongodb
from mongodb_backend.mongodb2 import mdbpool
from .erpwrapper import ErpWrapper
from yamlns import namespace as ns

from datetime import datetime
from plantmeter.resource import ProductionAggregator, ProductionPlant, ProductionMeter 
from plantmeter.mongotimecurve import MongoTimeCurve, toLocal, asUtc
from somutils.isodates import isodate, naiveisodatetime, localisodate

class GenerationkwhProductionAggregator(osv.osv):
    """
    Serialization class for a plant mix. See plantmeter.aggregator.
    A mix would be a set of plants which production is to be added together.
    """

    _name = 'generationkwh.production.aggregator'


    _columns = {
        'name': fields.char('Name', size=50),
        'description': fields.char('Description', size=150),
        'enabled': fields.boolean('Enabled'),
        'plants': fields.one2many('generationkwh.production.plant', 'aggr_id', 'Plants')
    }

    _defaults = {
        'enabled': lambda *a: False
    }

    def get_kwh(self, cursor, uid, mix_id, start, end, context=None):
        '''Get production aggregation'''
   
        if not context:
            context = {}
        _aggr = self._createAggregator(cursor, uid, mix_id)
        return _aggr.get_kwh(start, end).tolist()

    def firstActiveDate(self, cursor, uid, mix_id, context=None):
        '''Get first measurement date'''

        if not context:
            context = {}
        _aggr = self._createAggregator(cursor, uid, mix_id)
        date = _aggr.firstActiveDate()
        return date if date else None

    def firstMeasurementDate(self, cursor, uid, mix_id, context=None):
        '''Get first measurement date'''

        if not context:
            context = {}
        _aggr = self._createAggregator(cursor, uid, mix_id)
        date = _aggr.firstMeasurementDate()
        return date if date else None

    def lastMeasurementDate(self, cursor, uid, mix_id, context=None):
        '''Get last measurement date'''

        if not context:
            context = {}
        _aggr = self._createAggregator(cursor, uid, mix_id)
        date = _aggr.lastMeasurementDate()
        return date if date else None

    def _createAggregator(self, cursor, uid, mix_id):
        def extract_attrs(obj, attrs):
            # extracts name value tuples from erp browse object
            return {attr: getattr(obj, attr) for attr in attrs}

        if isinstance(mix_id, list) or isinstance(mix_id, tuple):
            mix_id = mix_id[0]

        aggr = self.browse(cursor, uid, mix_id)
        curveProvider = MongoTimeCurve(mdbpool.get_db(),
            'tm_profile',
            creationField = 'create_date',
            timestampField = 'utc_gkwh_timestamp',
            )

        # TODO: Clean initialization method
        args = ['id', 'name', 'description', 'enabled']
        return ProductionAggregator(**dict(
            extract_attrs(aggr, args),
            plants=[
                ProductionPlant(**dict(
                    extract_attrs(plant, args+['first_active_date']),
                    meters=[
                        ProductionMeter(
                            curveProvider=curveProvider,
                            **extract_attrs(meter, args + ['first_active_date'])
                            )
                        for meter in plant.meters
                        if meter.enabled
                        ],
                    ))
                for plant in aggr.plants
                if plant.enabled
                ],
            ))

GenerationkwhProductionAggregator()


class GenerationkwhProductionPlant(osv.osv):
    """
    Serialization class for a plant. See plantmeter.plant.
    """

    _name = 'generationkwh.production.plant'
    _columns = {
        'name': fields.char('Name', size=50),
        'description': fields.char('Description', size=150),
        'enabled': fields.boolean('Enabled'),
        'nshares': fields.integer('Number of shares'),
        'aggr_id': fields.many2one('generationkwh.production.aggregator', 'Production aggregator',
                                  required=True),
        'meters': fields.one2many('generationkwh.production.meter', 'plant_id', 'Meters'),
        'first_active_date': fields.date('First operative date'),
        'last_active_date': fields.date('Last operative date'),
    }
    _defaults = {
        'enabled': lambda *a: False,
    }


GenerationkwhProductionPlant()


class GenerationkwhProductionMeter(osv.osv):
    """
    Serialization class for a plant meter. See plantmeter.meter.
    """

    _name = 'generationkwh.production.meter'
    _columns = {
        'name': fields.char('Name', size=50),
        'description': fields.char('Description', size=150),
        'enabled': fields.boolean('Enabled'),
        'plant_id': fields.many2one('generationkwh.production.plant'),
        'first_active_date': fields.date('First operative date'),
        }
    _defaults = {
        'enabled': lambda *a: False,
    }

GenerationkwhProductionMeter()

class PlantShareProvider(ErpWrapper):
    """
    Provides a list for each plant of their active periods and their share value.
    To be used as provider for a LayeredShareCurve, that generates
    the curves to represent the share value of the active built plants.
    """
    def __init__(self, erp, cursor, uid, mixname, context=None):
        self.mixname = mixname
        super(PlantShareProvider, self).__init__(erp, cursor, uid, context)

    def items(self):
        Mix = self.erp.pool.get('generationkwh.production.aggregator')
        Plant = self.erp.pool.get('generationkwh.production.plant')
        mix_ids = Mix.search(self.cursor, self.uid, [
            ('name', '=', self.mixname)
        ])
        # TODO: if not mixids: ....

        plant_ids = Plant.search(self.cursor, self.uid, [
            ('aggr_id', '=', mix_ids[0]),
        ])
        plants = Plant.read(self.cursor, self.uid, plant_ids, [
        ])
        return [
            ns(
                mix=self.mixname,
                shares=plant['nshares'],
                firstEffectiveDate = isodate(plant['first_active_date']),
                lastEffectiveDate = isodate(plant['last_active_date']),
            )
            for plant in plants
        ]



class GenerationkwhProductionMeasurement(osv_mongodb.osv_mongodb):

    _name = 'generationkwh.production.measurement'
    _order = 'timestamp desc'

    _columns = {
        'name': fields.integer('Plant identifier'), # NOTE: workaround due mongodb backend
        'create_at': fields.datetime('Create datetime'),
        'datetime': fields.datetime('Exported datetime'),
        'daylight': fields.char('Exported datetime daylight',size=1),
        'ae': fields.float('Exported energy (kWh)')
    }

    def search(self, cursor, uid, args, offset=0, limit=0, order=None,
               context=None, count=False):
        '''Exact match for name.
        In mongodb, even when an index exists, not exact
        searches for fields scan all documents in collection
        making it extremely slow. Making name exact search
        reduces dramatically the amount of documents scanned'''

        new_args = []
        for arg in args:
            if not isinstance(arg, (list, tuple)):
                new_args.append(arg)
                continue
            field, operator, value = arg
            if field == 'name' and operator != '=':
                operator = '='
            new_args.append((field, operator, value))
        return super(GenerationkwhProductionMeasurement,
                     self).search(cursor, uid, new_args,
                                  offset=offset, limit=limit,
                                  order=order, context=context,
                                  count=count)

GenerationkwhProductionMeasurement()



class GenerationkwhProductionAggregatorTesthelper(osv.osv):
    '''Implements generationkwh production aggregation testhelper '''

    _name = 'generationkwh.production.aggregator.testhelper'
    _auto = False


    def get_kwh(self, cursor, uid, mix_id, start, end, context=None):
        mix = self.pool.get('generationkwh.production.aggregator')
        return mix.get_kwh(cursor, uid, mix_id,
                isodate(start), isodate(end), context)

    def firstActiveDate(self, cursor, uid, mix_id, context=None):
        mix = self.pool.get('generationkwh.production.aggregator')
        result = mix.firstActiveDate(cursor, uid, mix_id, context)
        return result and str(result)

    def firstMeasurementDate(self, cursor, uid, mix_id, context=None):
        mix = self.pool.get('generationkwh.production.aggregator')
        result = mix.firstMeasurementDate(cursor, uid, mix_id, context)
        return result and str(result)

    def lastMeasurementDate(self, cursor, uid, mix_id, context=None):
        production = self.pool.get('generationkwh.production.aggregator')
        result = production.lastMeasurementDate(cursor, uid, mix_id, context)
        return result and str(result)

    def clear_mongo_collections(self, cursor, uid, collections, context=None):
        for collection in collections:
            mdbpool.get_db().drop_collection(collection)

    def fillMeasurements(self, cursor, uid, first_date, meter_name, values):
        curveProvider = MongoTimeCurve(mdbpool.get_db(),
            'tm_profile',
            creationField = 'create_date',
            timestampField = 'utc_gkwh_timestamp',
            )
        curveProvider.update(
            start = localisodate(first_date),
            filter = meter_name,
            field = 'ae',
            data = values,
            )

    def fillMeasurementPoint(self, cursor, uid, pointTime, name, value, context=None):
        curveProvider = MongoTimeCurve(mdbpool.get_db(),
            'tm_profile',
            creationField = 'create_date',
            timestampField = 'utc_gkwh_timestamp',
            )
        curveProvider.fillPoint(
            datetime=toLocal(asUtc(naiveisodatetime(pointTime))),
            name=name,
            ae=value)

    def plantShareItems(self, cursor, uid, mixname):
        provider = PlantShareProvider(self, cursor, uid, mixname, context={})
        return [
            dict(
                item,
                lastEffectiveDate=str(item.lastEffectiveDate),
                firstEffectiveDate=str(item.firstEffectiveDate),
            )
            for item in provider.items()
        ]



GenerationkwhProductionAggregatorTesthelper()


# vim: ts=4 sw=4 et



#!/usr/bin/env python

import pymongo
from yamlns import namespace as ns
import numpy
import datetime

"""
+ More than one meassure
+ Different name ignored
+ Priority for newer meassuers the same hour
+ Summer daylight
+ Check mandatory fields
+ last meassure
+ first meassure
+ remove urlparse dependency on backends
+ Properly detect summer daylight change
- lastFullDate: proper implementation

"""

from somutils.isodates import (
    asUtc,
    toLocal,
    tz,
    addDays,
    addHours,
    )


hoursPerDay=25

"""
- Spain daylight:
    - UTC offset
        - +1h in winter
        - +2h in summer
    - Hourly curves have
        - an padding to the left in winter: _ X X X ... X X
        - an padding to the right in summer: X X X ... X _
    - Summer to winter is last sunday of october
        - that day has no padding in neither side
        - It has two 2:00h, a summer and a winter one
    - Winter to summer is last sunday of march
        - that day has paddings on both sides
        - It has no 2:00h, next to winter 1:00h is summer 3:00h
"""


def dateToCurveIndex(start, localTime):
    """
        Maps a timezoned datetime to a hourly curve index starting
        at 'start' date. Time part of 'start' is ignored.
        Both times' timezone should match curve's timezone.

        A day in houry curve has 25 positions. Padding is added
        to keep same solar time in the same position across summer
        daylight saving shift.
    """
    ndays = (localTime.date()-start.date()).days

    sofDay = toLocal(datetime.datetime(localTime.year,
                               localTime.month,
                               localTime.day))
    nextDay = tz.normalize(sofDay + datetime.timedelta(days=1))
    toWinterDls = sofDay.dst() and not nextDay.dst()
    toSummerDls = not sofDay.dst() and nextDay.dst()
    offset = 1 if toWinterDls and not localTime.dst() else (
             -1 if toSummerDls and localTime.dst() else 0)
    return localTime.hour + hoursPerDay*ndays + offset


def curveIndexToDate(start, index):
    """
        Maps an index withing an hourly curve starting at 'start' date
        into a local date.

        A day in houry curve has 25 positions. Padding is added
        to keep same solar time in the same position across summer
        daylight saving shift.
    """
    days = index//hoursPerDay
    localday = addDays(start, days)

    nextDay = tz.normalize(localday + datetime.timedelta(days=1))
    toWinterDls = localday.dst() and not nextDay.dst()
    toSummerDls = not localday.dst() and nextDay.dst()

    hours = index%hoursPerDay
    if hours == 24 and not toWinterDls: return None
    if hours == 23 and toSummerDls: return None
    return tz.normalize(localday+datetime.timedelta(hours=hours))



class MongoTimeCurve(object):
    """Consolidates curve data in a mongo database (old format)"""

    def __init__(self, mongodb, collection,
            timestampField='datetime',
            creationField='create_at',
        ):
        self.db = mongodb
        self.collectionName = collection
        self.collection = self.db[collection]
        self.timestamp = timestampField
        self.creation = creationField

    def get(self, start, stop, filter, field, filling=None):
        assert start.tzinfo is not None, (
            "MongoTimeCurve.get called with naive (no timezone) start date")

        assert stop.tzinfo is not None, (
            "MongoTimeCurve.get called with naive (no timezone) stop date")

        ndays = (stop.date()-start.date()).days+1
        data = numpy.zeros(ndays*hoursPerDay, int)
        if filling :
            filldata = numpy.zeros(ndays*hoursPerDay, bool)
        filters = {
            self.timestamp: {
                '$gte': start,
                '$lt': addDays(stop,1)
            }
        }
        if isinstance(filter, dict):
            filters.update(filter)
        elif filter: filters.update(name=filter)
        from bson.son import SON

        pipeline = [
            # pick the ones matching the filters
            {"$match":
                filters,
            },
            # sort by timestamp, name and new firsts
            {"$sort": SON([
                (self.timestamp, 1),
                ("name", 1),
                (self.creation, -1),
            ])},
            # group all having the same timestamp and name, pick newest on collission
            {"$group": {
                '_id': {
                    self.timestamp: '$'+self.timestamp,
                    'name': '$name',
                },
                self.timestamp: {'$first': '$'+self.timestamp},
                field: {'$first': '$'+field}
            }},
            # Suma tots els que tenen el mateix timestamp, diferent nom
            {"$group": {
                '_id': '$'+self.timestamp,
                self.timestamp: {'$first': '$'+self.timestamp},
                field: {'$sum': '$'+field},
            }},
        ]

        for x in self.collection.aggregate(pipeline,cursor={},allowDiskUse=True):
            point = ns(x)
            localTime = toLocal(asUtc(point[self.timestamp]))
            timeindex = dateToCurveIndex (start, localTime)
            data[timeindex]=point.get(field)
            if filling: filldata[timeindex]=True

        if filling: return data, filldata
        return data

    def fillPoint(self, **data):
        for requiredField in ('name', 'datetime'):
            if requiredField not in data:
                raise Exception("Missing '{}'".format(requiredField))

        assert data['datetime'].tzinfo is not None, (
            "MongoTimeCurve.fillPoint with naive (no timezone) datetime")

        counter = self.db['counters'].find_and_modify(
            {'_id': self.collectionName},
            {'$inc': {'counter': 1}}
        )
        timestamp = data.pop('datetime')
        data.update({
            self.creation: datetime.datetime.now(),
            self.timestamp: timestamp,
            })
        return self.collection.insert(data)

    def _firstLastDate(self, name, first=False):
        """returns the date of the first or last item of a given name"""
        order = pymongo.ASCENDING if first else pymongo.DESCENDING
        for point in (self.collection
                .find(dict(name=name))
                .sort(self.timestamp, order)
                .limit(1)
                ):
            return toLocal(asUtc(point[self.timestamp])).replace(
                    hour=0,minute=0,second=0)
        return None

    def firstDate(self, name):
        """returns the date of the first item of a given name"""
        return self._firstLastDate(name, first=True)

    def lastDate(self, name):
        """returns the date of the last item of a given name"""
        return self._firstLastDate(name)

    def lastFullDate(self,name):
        order = pymongo.DESCENDING
        for point in (self.collection
                .find(dict(name=name))
                .sort(self.timestamp, order)
                .limit(1)
                ):
            return toLocal(addHours(asUtc(point[self.timestamp]),-18)).replace(
                    hour=0,minute=0,second=0)
    def firstFullDate(self,name):
        # TODO: dumb implementation, having just a single hour considers whole date filled
        return self.firstDate(name)

    def update(self, start, filter, field, data):
        """Updates the curve with new data"""

        assert start.tzinfo is not None, (
            "MongoTimeCurve.update called with naive (no timezone) start date")

        if isinstance(filter, str) or isinstance(filter, int):
            filter = dict(name=filter)

        stop = start + datetime.timedelta(days=len(data)//hoursPerDay+1)
        oldData, filling = self.get(start, stop, filter, field, filling=True)
        if type(data) == numpy.ndarray:
            data = (x.item() for x in data)
        for i,(bin,old,f) in enumerate(zip(data,oldData,filling)):
            curveDate = curveIndexToDate(start, i)
            if curveDate is None: continue
            if bin == old: continue
            self.fillPoint(
                datetime=curveDate,
                name=filter['name'],
                **{field: bin}
                )



# vim: et ts=4 sw=4



#!/usr/bin/env python
from setuptools import setup, find_packages
import sys

py2 = sys.version_info<(3,)

with open('requirements.txt') as f:
    INSTALL_REQUIRES = [x.strip() for x in f.readlines()]

with open('requirements-dev.txt') as f:
    TEST_REQUIRES = [x.strip() for x in f.readlines()]

with open('README.md') as f:
    readme = f.read()

setup(
    name = "plantmeter",
    version = "1.7.10",
    description =
        "OpenERP module and library to manage multisite energy generation",
    author = "Som Energia SCCL",
    author_email = "info@somenergia.coop",
    url = 'https://github.com/Som-Energia/plantmeter',
    long_description = readme,
    long_description_content_type = 'text/markdown',
    license = 'GNU General Public License v3 or later (GPLv3+)',
    packages=find_packages(exclude=['*[tT]est*']),
    include_package_data = True,
    install_requires=INSTALL_REQUIRES,
    setup_requires=["pytest-runner"],
    tests_require=TEST_REQUIRES,
    test_suite = 'plantmeter',
    classifiers = [
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Intended Audience :: Developers',
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
    ],
)

